From f0b21250875726bd429bb9657b4037be41f745e4 Mon Sep 17 00:00:00 2001
From: Ubuntu <ubuntu@ip-172-31-89-33.ec2.internal>
Date: Thu, 11 Jul 2024 15:05:37 +0000
Subject: [PATCH 1/1] vaccum_delay with absolute time nanosleep

---
 src/backend/commands/vacuum.c | 49 ++++++++++++++++++++++++++++++++++-
 1 file changed, 48 insertions(+), 1 deletion(-)

diff --git a/src/backend/commands/vacuum.c b/src/backend/commands/vacuum.c
index 48f8eab202..a65295b2ff 100644
--- a/src/backend/commands/vacuum.c
+++ b/src/backend/commands/vacuum.c
@@ -116,6 +116,53 @@ static bool vacuum_rel(Oid relid, RangeVar *relation, VacuumParams *params,
 static double compute_parallel_delay(void);
 static VacOptValue get_vacoptval_from_boolean(DefElem *def);
 static bool vac_tid_reaped(ItemPointer itemptr, void *state);
+static void vacuum_sleep(double msec);
+
+static
+void vacuum_sleep(double msec)
+{
+	long microsec = msec * 1000;
+
+	if (microsec > 0)
+	{
+#ifndef WIN32
+		/*
+		 * We allow nanosleep to handle interrupts and retry with the remaining time.
+		 * However, since nanosleep is susceptible to time drift when interrupted
+		 * frequently, we add a safeguard to break out of the nanosleep whenever the
+		 * total time of the sleep exceeds the requested sleep time. Using nanosleep
+		 * is a more portable approach than clock_nanosleep.
+		 */
+		struct timespec delay;
+		struct timespec remain;
+		struct instr_time start_time;
+
+		INSTR_TIME_SET_CURRENT(start_time);
+
+		delay.tv_sec = microsec / 1000000L;
+		delay.tv_nsec = (microsec % 1000000L) * 1000;
+
+		while(nanosleep(&delay, &remain) == -1 && errno == EINTR)
+		{
+			struct instr_time current_time;
+			struct instr_time elapsed_time;
+
+			INSTR_TIME_SET_CURRENT(current_time);
+
+			INSTR_TIME_SET_ZERO(elapsed_time);
+			INSTR_TIME_ACCUM_DIFF(elapsed_time, current_time, start_time);
+
+			if (INSTR_TIME_GET_MICROSEC(elapsed_time) >= microsec)
+				break;
+
+			delay = remain;
+		}
+
+#else
+		SleepEx((microsec < 500 ? 1 : (microsec + 500) / 1000), FALSE);
+#endif
+	}
+}
 
 /*
  * GUC check function to ensure GUC value specified is within the allowable
@@ -2384,7 +2431,7 @@ vacuum_delay_point(void)
 			msec = vacuum_cost_delay * 4;
 
 		pgstat_report_wait_start(WAIT_EVENT_VACUUM_DELAY);
-		pg_usleep(msec * 1000);
+		vacuum_sleep(msec);
 		pgstat_report_wait_end();
 
 		/*
-- 
2.43.0

